
meta_field_tool = [
    {'id': 2, 'sort': 32, 'pid': 7, 'isLeaf': True, 'name': '输入框', 'code': 'text'},
    {'id': 3, 'sort': 17, 'pid': 0, 'isLeaf': False, 'name': 'SPEC', 'code': '特殊工具集'},
    {'id': 5, 'sort': 22, 'pid': 0, 'isLeaf': False, 'name': 'DATE', 'code': '日期时间工具集'},
    {'id': 6, 'sort': 21, 'pid': 8, 'isLeaf': True, 'name': '数字输入', 'code': 'number'},
    {'id': 7, 'sort': 36, 'pid': 0, 'isLeaf': False, 'name': 'TEXT', 'code': '文本工具集'},
    {'id': 8, 'sort': 26, 'pid': 0, 'isLeaf': False, 'name': 'NUMB', 'code': '数字工具集'},
    {'id': 9, 'sort': 10, 'pid': 0, 'isLeaf': False, 'name': 'FILE', 'code': '文件工具集'},
    {'id': 10, 'sort': 27, 'pid': 0, 'isLeaf': False, 'name': 'BOOL', 'code': '逻辑工具集'},
    {'id': 11, 'sort': 31, 'pid': 0, 'isLeaf': False, 'name': 'LIST', 'code': '列表工具集'},
    {'id': 12, 'sort': 8, 'pid': 3, 'isLeaf': True, 'name': '对象', 'code': 'object'},
    {'id': 13, 'sort': 5, 'pid': 9, 'isLeaf': True, 'name': '图片上传', 'code': 'image'},
    {'id': 14, 'sort': 2, 'pid': 9, 'isLeaf': True, 'name': '文件上传', 'code': 'file'},
    {'id': 15, 'sort': 13, 'pid': 9, 'isLeaf': True, 'name': '富文本', 'code': 'rich'},
    {'id': 17, 'sort': 11, 'pid': 10, 'isLeaf': True, 'name': '开关', 'code': 'switch'},
    {'id': 18, 'sort': 7, 'pid': 3, 'isLeaf': True, 'name': '元数据', 'code': 'meta'},
    {'id': 19, 'sort': 9, 'pid': 7, 'isLeaf': True, 'name': '颜色选择', 'code': 'color'},
    {'id': 20, 'sort': 14, 'pid': 11, 'isLeaf': True, 'name': '穿梭框', 'code': 'transfer'},
    {'id': 21, 'sort': 16, 'pid': 7, 'isLeaf': True, 'name': '自动填充', 'code': 'auto'},
    {'id': 22, 'sort': 35, 'pid': 5, 'isLeaf': True, 'name': '日期选择', 'code': 'date'},
    {'id': 23, 'sort': 12, 'pid': 10, 'isLeaf': True, 'name': '逻辑选择', 'code': 'boolean'},
    {'id': 24, 'sort': 24, 'pid': 11, 'isLeaf': True, 'name': '列表选择', 'code': 'select'},
    {'id': 25, 'sort': 15, 'pid': 11, 'isLeaf': True, 'name': '树形选择', 'code': 'tree'},
    {'id': 26, 'sort': 23, 'pid': 11, 'isLeaf': True, 'name': '及联选择', 'code': 'cascade'},
    {'id': 28, 'sort': 25, 'pid': 7, 'isLeaf': True, 'name': '图标', 'code': 'icon'},
    {'id': 31, 'sort': 6, 'pid': 0, 'isLeaf': True, 'name': '无', 'code': 'none'},
    {'id': 32, 'sort': 30, 'pid': 7, 'isLeaf': True, 'name': '文本框', 'code': 'textarea'},
    {'id': 33, 'sort': 18, 'pid': 36, 'isLeaf': True, 'name': '时间区间', 'code': 'timerange'},
    {'id': 35, 'sort': 33, 'pid': 5, 'isLeaf': True, 'name': '时间选择', 'code': 'time'},
    {'id': 36, 'sort': 20, 'pid': 0, 'isLeaf': False, 'name': 'RANGE', 'code': '区间工具集'},
    {'id': 37, 'sort': 38, 'pid': 36, 'isLeaf': True, 'name': '日期区间', 'code': 'daterange'},
    {'id': 39, 'sort': 3, 'pid': 36, 'isLeaf': True, 'name': '多日期', 'code': 'dates'},
    {'id': 54, 'sort': 54, 'pid': 7, 'isLeaf': True, 'name': '集合', 'code': 'set'}
]

meta_field_domain = [
    {
        'name': 'CharField',
        'default_id': 'text', 'align': 'left',
        'tools': [
            'text', 'number', 'meta', 'color', 'auto', 'date', 'select', 'tree', 'cascade', 'icon',
            'textarea', 'timerange', 'daterange', 'dates', 'set'
        ]
     },
    {
        'name': 'TextField',
        'default_id': 'textarea', 'align': 'left',
        'tools': ['text', 'textarea']
    },
    {
        'name': 'BooleanField',
        'default_id': 'switch', 'align': 'center',
        'tools': ['switch', 'boolean']
    },
    {
        'name': 'IntegerField',
        'default_id': 'number', 'align': 'right',
        'tools': ['number']
    },
    {
        'name': 'FloatField',
        'default_id': 'number', 'align': 'right',
        'tools': ['number']
    },
    {
        'name': 'ForeignKey',
        'default_id': 'select', 'align': 'left',
        'tools': ['select', 'tree', 'cascade']
    },
    {
        'name': 'ManyToOneRel',
        'default_id': 'select', 'align': 'center',
        'tools': ['transfer', 'select', 'tree', 'cascade']
    },
    {
        'name': 'ManyToManyField',
        'default_id': 'select', 'align': 'center',
        'tools': [ 'transfer', 'select', 'tree', 'cascade']
    },
    {
        'name': 'ManyToManyRel',
        'default_id': 'select', 'align': 'center',
        'tools': ['transfer', 'select', 'tree', 'cascade']
    },
    {
        'name': 'OneToOneRel',
        'default_id': 'none', 'align': 'left',
        'tools': []
    },
    {
        'name': 'OneToOneField',
        'default_id': 'none', 'align': 'left',
        'tools': []
    },
    {
        'name': 'DateField',
        'default_id': 'date', 'align': 'center',
        'tools': ['date']
    },
    {
        'name': 'TimeField',
        'default_id': 'time', 'align': 'center',
        'tools': ['time']
    },
    {
        'name': 'DateTimeField',
        'default_id': 'date', 'align': 'center',
        'tools': ['date']
    },
    {
        'name': 'JSONField',
        'default_id': 'object', 'align': 'center',
        'tools': ['object']
    },
    {
        'name': 'FileField',
        'default_id': 'file', 'align': 'center',
        'tools': ['image', 'file', 'rich']
    },
    {
        'name': 'BigAutoField',
        'default_id': 'none', 'align': 'right',
        'tools': []
    },
    {
        'name': 'UUIDField',
        'default_id': 'none', 'align': 'left',
        'tools': []
    },
    {
        'name': 'Custom',
        'default_id': 'none', 'align': 'left',
        'tools': []
    },
]




from ..singleton_meta import SingletonMeta
from .engine_mon import MonEngine
from .engine_orm import OrmEngine


class DaoEngine(metaclass=SingletonMeta):

    def __init__(self):
        self.mon = MonEngine()
        self.orm = OrmEngine()




from ...channels.sender import ValarSocketSender
from ...core.counter import Counter
from ...core.dao.dao_base import Dao
from ...core.dao.dao_mon import MonDao
from ...core.dao.dao_orm import OrmDao




def save_many_handler(sender: ValarSocketSender):
    data = sender.data
    entity, array, db = data.get("entity"), data.get("array",[]), data.get("db")
    dao = Dao(entity, db)
    counter = Counter(array)
    for item in array:
        item['saved'] = True
        dao.save_one(item)
        payload = counter.tick()
        sender.to_clients(payload, sender.client, wait=True)



def delete_many_handler(sender: ValarSocketSender):
    data = sender.data
    entity, conditions, db = data.get("entity"), data.get("conditions", []), data.get("db")
    if db == 'orm':
        dao = OrmDao(entity)
        query_set, _ = dao.find(conditions)
        paths = dao.model.get_file_paths(query_set)
        query_set.delete()
        counter = Counter(len(paths))
        for path in paths:
            dao.model.minio.remove_path(path)
            payload = counter.tick()
            sender.to_clients(payload, sender.client, wait=True)
    else:
        dao = MonDao(entity)
        dao.delete(conditions)







def __translate_finder__(conditions):
    return {}


class MonQuery:

    def __init__(self, conditions: list, orders = None):
        self.orders = orders or {'sort': -1}
        conditions = conditions if len(conditions) else [{'includes':{},'excludes':{}}]
        self.finder = __translate_finder__(conditions)




from bson import ObjectId
from bson.errors import InvalidId
from pymongo.synchronous.collection import Collection


class MonModel:

    def __init__(self, database, entity):
        self.entity = entity
        self.name = entity.replace('.', '_')
        self.manager: Collection = database[self.name]

    @staticmethod
    def object_id(_id):
        try:
            return ObjectId(_id)
        except(InvalidId, TypeError):
            return None

    def detach_item(self, item):
        _id = item.get('id')
        if _id:
            del item['id']
        return self.object_id(_id), item


import datetime
from django.core.paginator import Paginator
from django.db.models import QuerySet

from .engine import DaoEngine
from .query_orm import OrmQuery
from ..dao_abstract import AbstractDao


class OrmDao(AbstractDao):
    def __init__(self, entity):
        self.entity = entity
        engine = DaoEngine().orm
        self.model = engine.get_model(entity)
        self.objects = self.model.manager

    def get_model(self):
        return self.model

    def save_one(self, item):
        oid, simple_item, complex_item  = self.model.detach_item(item)
        query_set = self.objects.filter(id=oid) if oid else []
        if len(query_set):
            simple_item['modify_time'] = datetime.datetime.now()
            query_set.update(**simple_item)
            bean = query_set.first()
        else:
            bean = self.objects.create(**simple_item)
            bean.sort = bean.id
            bean.save()
        self.model.save_complex_field(complex_item, bean)
        bean.save()
        return bean

    def delete_one(self, _id):
        oid = self.model.object_id(_id)
        flag = False
        if oid:
            query_set = self.objects.filter(id=oid)
            self.model.remove_files(query_set)
            query_set.delete()
            flag = True
        return flag

    def find_one(self, _id):
        oid = self.model.object_id(_id)
        return self.objects.filter(id=oid).first() if oid else None



    def find(self, conditions=None, orders=None,size=0, page=1):
        query = OrmQuery(conditions, orders)
        query_set = self.objects.filter(query.includes).exclude(query.excludes).order_by(*query.orders)
        total = query_set.count()
        if size:
            paginator = Paginator(query_set, size)
            query_set = paginator.page(page).object_list
        return query_set, total

    def update(self, template, conditions):
        if template and len(template.keys()):
            oid, simple_item, complex_item = self.model.detach_item(template)
            query_set, total = self.find(conditions)
            query_set.update(**simple_item)
            return True
        return False

    def delete(self, conditions):
        query_set, total = self.find(conditions)
        self.model.remove_files(query_set)
        query_set.delete()

    def transform(self, o, code=None):
        if isinstance(o, QuerySet):
            return self.model.to_dict(o, code)
        else:
            return o.full()

    def tree(self, root, conditions=None):
        all_set, _ = self.find([])
        query = OrmQuery(conditions)
        if query.is_empty(conditions):
            return all_set
        values = all_set.values('id', 'pid')
        mapping = {item['id']: item['pid'] for item in values}
        results, _ = self.find(conditions)
        id_set = {root}
        for item in results:
            _id = item.id
            route = []
            while _id is not None:
                route.append(_id)
                _id = mapping.get(_id)
            if root in route:
                id_set.update(route)
        return all_set.filter(id__in=id_set).order_by('-sort')



__mf_common__ = ['prop','name']

meta_field_key_defaults = {
    'data.Meta': {
        'default': ('pick', ['entity','name']),
    },
    'data.MetaView': {
        'list': ('pick', ['meta_id','code','view_name']),
    },
    'data.MetaField': {
        'add': ('pick',['prop','domain','name']),
        'tool': ('pick',[*__mf_common__,'domain','tool','refer','format']),
        'rest': ('pick',[*__mf_common__,'not_null','allow_edit','allow_sort','allow_search','allow_download','allow_upload','allow_update']),
        'table': ('pick',[*__mf_common__,'unit','column_width','fixed','align','edit_on_table','hide_on_table','header_color','cell_color']),
        'form': ('pick',[*__mf_common__,'hide_on_form','hide_on_form_insert','hide_on_form_edit','hide_on_form_branch','hide_on_form_leaf','span']),
    }
}



meta_field_value_defaults = {
    'data.MetaFieldDomain':{
        "__init__":{
            "default_id":{
                "tool":"tree"
            },
            "search_id":{
                "tool":"tree"
            },
            "tools":{
                "tool":"tree",
                "refer": {
                    "display":"code"
                }
            },
            "align":{
                "tool":"set",
                "format":{
                    "set": {
                        'left':'左对齐',
                        'right':'右对齐',
                        'center':'剧中对齐',
                    }
                }
            }
        }
    },
    'data.MetaField':{
        "__init__":{
            "column_width":{
                'unit':'px'
            },
            "fixed":{
                "tool":"set",
                "format":{
                    "set": {
                        'left':'左侧固定',
                        'right':'右侧固定',
                    }
                }
            },
            "align":{
                "tool":"set",
                "format":{
                    "set": {
                        'left':'左对齐',
                        'right':'右对齐',
                        'center':'剧中对齐',
                    }
                }
            },
            "prop":{
                'allow_edit': False,
                'column_width': 120
            },
            "domain":{
                'allow_edit': False,
                'column_width': 120,
            },
            "tool":{
                'column_width': 100,
                'tool': 'tree',
                'refer': {
                    'entity':'data.MetaFieldTool',
                    'includes': {'metafielddomain__name':'$domain'},
                    'value': 'code'
                }

            },
            "span":{
                'column_width': 100,
                "format": { "min": 0, "max": 24, "step": 1, "precision": 0, "step_strictly": True }
            },
            "refer":{
                'allow_edit': False,
                'column_width': 80
            },
            "format":{
                'allow_edit': False,
                'column_width': 80
            },
        }
    }
}


import pymongo
from django.conf import settings

from .model_mon import MonModel


class MonEngine:

    def __init__(self):
        uri = f'mongodb://localhost:27017/'
        mongo = settings.MONGO
        if mongo:
            param = ['host', 'port', 'username', 'password']
            host, port, username, password = [mongo.get(p) for p in param]
            uri = f'mongodb://{username}:{password}@{host}:{port}/'

        client = pymongo.MongoClient(uri, **{
            'maxPoolSize': 10,
            'minPoolSize': 0,
            'maxIdleTimeMS': 10000,
            'connectTimeoutMS': 10000,
            'socketTimeoutMS': 10000,
            'serverSelectionTimeoutMS': 10000,
        })
        database =  client[settings.BASE_APP]
        self.uri = uri
        self.client = client
        self.database = database

    def get_mapping(self):
        return  {col['name']: self.database[col['name']] for col in self.database.list_collections()}

    def get_model(self,entity)->MonModel:
        return MonModel(self.database, entity)


from deepmerge import always_merger
from django.db.models import ManyToOneRel, ForeignKey, ManyToManyRel, ManyToManyField,OneToOneField,OneToOneRel
from django.db.models import IntegerField, BooleanField, FloatField,TextField, CharField
from django.db.models import FileField, JSONField
from django.db.models import DateTimeField, TimeField, DateField



class OrmField:

    def __init__(self, entity, field, is_tree):
        self.entity = entity
        self.model_field = field
        self.is_tree = is_tree
        self.not_null = not field.null
        self.clazz = type(field)
        self.multiple = self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel]

        self.prop = self.__prop__()
        self.domain = self.__domain__()
        self.model = self.__model__()
        self.label = self.__label__()
        self.column_width = self.__column_width__()
        self.refer = self.__refer__()
        self.align = self.___align__()
        self.format = self.__formating__()





    def json(self):
        _field = {
            "prop": self.prop,
            "label": self.label,
            "name": self.label,
            "domain": self.domain,
            "refer": self.refer,
            "format": self.format,
            "not_null": self.not_null,
            "align": self.align,
            "column_width": self.column_width,
        }
        if self.is_tree:
            if self.prop in ['pid', 'isLeaf']:
                _field['hide_on_table'] = True
                _field['hide_on_form'] = True
                _field['hide_on_form_branch'] = True
                _field['hide_on_form_leaf'] = True
            elif self.prop in ['icon']:
                _field['tool'] = 'icon'
        return _field


    def __column_width__(self):
        if self.clazz in [BooleanField, FileField, JSONField]:
            return 100
        elif self.clazz in [DateField, DateTimeField, TimeField]:
            return 120
        return 0

    def __label__(self):
        return self.model._meta.verbose_name \
            if self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel, OneToOneRel, OneToOneField] \
            else self.model_field.verbose_name


    def __model__(self):
        return self.model_field.related_model \
            if self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel, OneToOneRel, OneToOneField, ForeignKey] \
            else None


    def __domain__(self):
        return self.clazz.__name__ \
            if self.clazz in [ManyToOneRel, ManyToManyField, ManyToManyRel ,OneToOneRel, OneToOneField] \
            else self.model_field.get_internal_type()

    def __prop__(self):
        return self.model_field.name + "_id" \
            if self.clazz in [ForeignKey,OneToOneRel, OneToOneField] \
            else self.model_field.name


    def ___align__(self):
        if self.clazz in [FloatField, IntegerField]:
            return 'right'
        elif self.clazz in [BooleanField, FileField, JSONField, DateField, DateTimeField, TimeField]:
            return 'center'
        return 'left'

    def __refer__(self):
        refer = {
            "entity": None,
            "value": "name", "label": 'name', "display": "id",
            "multiple": self.multiple, "strict": False, "remote": False,
            "includes": {}, "excludes": {},
            "root": 0, "isTree": self.is_tree
        }
        if self.model:
            module, name = self.model.__module__, self.model.__name__
            refer['entity'] = '%s.%s' % (module.replace('.models', '').split('.')[-1], name)
        return refer

    def __formating__(self):
        _format = {
            # 文本
            "maxlength": 0,
            "type": 'text',

            # 数值
            "min": None,
            "max": None,
            "step": 1,
            "precision": None,
            "step_strictly": False,

            # 日期
            "frequency": "date",

            # 文件
            "maximum": 5,
            "accept": [],
            "width": 800,
            "height": 0,
            "file_name_field":None,
            "locked": False,

            #集合
            "set": {}
        }
        if self.clazz == CharField:
            _format['maxlength'] = self.model_field.max_length
        if self.clazz == TextField:
            _format['type'] = "textarea"
        elif self.clazz == DateTimeField:
            _format['frequency'] = "datetime"
        elif self.clazz == IntegerField:
            _format['precision'] = 0
            _format['step_strictly'] = True
        return _format






from django.db.models import Q
from functools import reduce



def __fun__(x, y): return x | y
def __translate_orders__(orders):
    array = []
    for key in orders:
        value = orders.get(key)
        prefix = '-' if value == -1 else ''
        array.append(f'{prefix}{key}')
    return array
def __translate_condition__(conditions, _type):
    


    return reduce(__fun__,  [Q(**cond[_type]) for cond in conditions])


class OrmQuery:

    def __init__(self, conditions: list, orders = None):
        self.orders =  __translate_orders__(orders or {'sort': -1})
        conditions = conditions if len(conditions) else [{'includes':{},'excludes':{}}]
        self.includes = __translate_condition__(conditions,'includes')
        self.excludes = __translate_condition__(conditions, 'excludes')

    @staticmethod
    def is_empty(conditions):
        if conditions and len(conditions):
            temp = {}
            for cond in conditions:
                temp.update(cond)
            return len(temp.keys()) >0
        else:
            return True









from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin
from response import ValarResponse


class Middleware(MiddlewareMixin):
    @staticmethod
    def process_response(request: HttpRequest, response: ValarResponse):
        headers = response.headers
        if type(response)==ValarResponse:
            message, code = response.message, response.code
            headers['valar_message'] = message
            headers['valar_code'] = code

        keys = ['client','auth','uid']
        for key in keys:
            value = request.headers.get(key)
            if value:
                headers[key] = value
        return response


from deepmerge import always_merger
from django.db.models import ManyToOneRel, ForeignKey, ManyToManyRel, ManyToManyField, OneToOneField, OneToOneRel
from django.db.models import QuerySet
from django.db.models import Manager
from django.db.models.fields.files import FieldFile
from django.forms import FileField

from .engine_minio import MinioEngine
from .utils_orm import OrmUtils
from ..meta.defaults.field_keys_default import meta_field_key_defaults
from ..meta.defaults.field_values_default import meta_field_value_defaults
from ..meta.field_orm import OrmField
from ..meta.init_meta_frame import init_meta_frame
from ..valar_models import VModel, VTree
from ...data.models import  MetaFieldDomain


class OrmModel:

    def __init__(self, mod, entity):
        self.entity = entity
        self.model = mod
        meta = getattr(mod, '_meta')
        self.name = meta.verbose_name
        self.is_tree = issubclass(mod, VTree)
        self.manager: Manager = mod.objects
        fields = meta.get_fields()
        mapping = {}
        for f in fields:
            field = OrmField(entity, f, self.is_tree)
            mapping[field.prop] = field
        self.mapping = mapping
        self.minio = MinioEngine()
        self.bucket_name = self.minio.get_bucket_name(entity)

    @staticmethod
    def object_id(_id):
        return _id if isinstance(_id, int) and _id else None

    def props(self, domain=None):
        array = []
        for prop in self.mapping:
            field: OrmField = self.mapping[prop]
            if field.domain == domain or domain is None:
                array.append(prop)
        return array


    def meta_frame(self):
        query_set = MetaFieldDomain.objects.all()
        if query_set.count() == 0:
            query_set = init_meta_frame()
        values = query_set.values('name', 'default__code', 'align')
        return {
            vs['name']: {
                "tool": vs['default__code'],
                "align": vs['align'],
            }
            for vs in values
        }




    def initial_fields(self, code):
        props = self.props()
        default_keys = meta_field_key_defaults.get(self.entity,{})
        method, array = default_keys.get(code, ('omit',[]))
        def fun(prop): return prop not in array if method == 'omit' else prop in array
        props = [prop for prop in props if fun(prop)]

        default_values = meta_field_value_defaults.get(self.entity, {})
        init_values = default_values.get('__init__', {})
        code_values = default_values.get(code, {})
        default_fields = always_merger.merge(init_values, code_values)

        meta_frame = self.meta_frame()
        fields = []
        for prop in props:
            field = self.get_field(prop)
            field_json = field.json()
            frame = meta_frame.get(field.domain, {})
            field_json.update(frame)
            default_field = default_fields.get(prop, {})
            always_merger.merge(field_json, default_field)
            fields.append(field_json)

        fields.reverse()
        return fields

    def get_field(self, prop)->OrmField:
        return self.mapping[prop]

    def detach_item(self, item):
        _id = item.get('id')
        if _id:
            del item['id']
        simple_item = {}
        complex_item = {}
        for prop in item:
            field = self.get_field(prop)
            value = item.get(prop)
            if field.domain in ['ManyToOneRel', 'ManyToManyField', 'ManyToManyRel', 'OneToOneRel', 'OneToOneField','FileField']:
                complex_item[prop] = value
            else:
                simple_item[prop] = value
        return self.object_id(_id), simple_item, complex_item

    def get_file_paths(self, query_set: QuerySet):
        props = self.props('FileField')
        items = query_set.values(*props)
        array = []
        for item in items:
            array += [i for i in item.values() if i]
        return array

    def remove_files(self, query_set: QuerySet):
        paths = self.get_file_paths(query_set)
        for path in paths:
            self.minio.remove_path(path)

    def save_complex_field(self, complex_item, bean):
        for prop in complex_item:
            value = complex_item[prop]
            field = self.get_field(prop).model_field
            clazz = type(field)
            if clazz == ManyToManyField:
                m2m = getattr(bean, prop)
                m2m.clear()
                m2m.add(*value)
            elif clazz == ManyToOneRel:
                getattr(bean, field.get_accessor_name()).clear()
                remote_model: VModel = field.related_model
                new_set: QuerySet = remote_model.objects.filter(id__in=value)
                remote_field: ForeignKey = field.remote_field
                k = remote_field.get_attname()
                new_set.update(**{k: bean.id})
            elif clazz == ManyToManyRel:
                getattr(bean, field.get_accessor_name()).clear()
                remote_model: VModel = field.related_model
                remote_items: QuerySet = remote_model.objects.filter(id__in=value)
                remote_field: ManyToManyField = field.remote_field
                remote_field_prop = remote_field.get_attname()
                for _bean in remote_items:
                    bean_set = getattr(_bean, remote_field_prop)
                    bean_set.add(bean)
            elif clazz == OneToOneRel:
                remote_model: VModel = field.related_model
                remote_field: OneToOneField = field.remote_field
                remote_field_prop = remote_field.get_attname()
                _bean = remote_model.objects.get(id=value)
                __bean = remote_model.objects.filter(**{remote_field_prop: bean.id}).first()
                if __bean:
                    setattr(__bean, remote_field_prop, None)
                    __bean.save()
                setattr(_bean, remote_field_prop, bean.id)
                _bean.save()
            elif clazz == OneToOneField:
                __bean = field.model.objects.filter(**{prop: value}).first()
                if __bean:
                    setattr(__bean, prop, None)
                    __bean.save()
                setattr(bean, prop, value)
            elif clazz == FileField:
                file_name, _bytes = value
                field_file: FieldFile = getattr(bean, prop)
                if field_file:
                    path = field_file.name
                    self.minio.remove_path(path)
                object_name = self.minio.get_object_name(bean.id, prop, file_name)
                path = self.minio.upload(self.bucket_name, object_name, _bytes) if _bytes else None
                setattr(bean, prop, path)




    def to_dict(self, query_set: QuerySet, code=None):
        # query_set = query_set.filter(saved=True)
        utils = OrmUtils()
        orm_fields = self.mapping.values()
        # 简单字段取值
        simple_props = [field.prop for field in orm_fields if field.domain not in utils.referred_domains]
        custom_props = utils.custom_props(self.entity, code)
        results = list(query_set.filter().values(*[*simple_props, *custom_props]))
        utils.date_values(orm_fields, results)
        # 关系型字段取值
        mapping = { row['id']: row for row in results}
        referred_fields = [field for field in orm_fields if field.domain in utils.referred_domains]
        pks = mapping.keys()
        for field in referred_fields:
            manager: Manager = query_set.model.objects
            qs = manager.filter(id__in=pks)
            utils.linkage(field, qs, mapping)
        return results




def array2tree(data, mapping):
    mapping = mapping or {}
    lookup = {}
    for array in data:
        for i in range(len(array)):
            key = '/'.join(array[0:i+1])
            item = mapping.get(key, {})
            value = item.get('value', array[i])
            label = item.get('label', value)
            display = item.get('display')
            item = lookup.get(key, {'value': value,'label':label,'display':display})
            if i < len(array) -1:
                item['children'] = item.get('children', [])
            lookup[key] = item
            if i > 0:
                parent =   '/'.join(array[0:i])
                lookup[parent]['children'].append(lookup[key])
    return [lookup[root] for root in [*set([array[0] for array in data])]]


from .dao_mon import MonDao
from .dao_orm import OrmDao
from ..dao_abstract import AbstractDao


class Dao(AbstractDao):

    def __init__(self, entity, db='orm'):
        self.entity = entity
        self.db = db
        self.dao: AbstractDao = OrmDao(entity) if db == 'orm' else MonDao(entity)

    def get_model(self):
        self.dao.get_model()

    def save_one(self, item):
        return self.dao.save_one(item)

    def delete_one(self, _id):
        return self.dao.delete_one(_id)

    def find_one(self, _id):
        return self.dao.find_one(_id)

    def find(self, conditions=None, orders=None, size=0, page=1):
        return self.dao.find(conditions, orders, size, page)

    def update(self, template, conditions):
        return self.dao.update(template, conditions)

    def delete(self, conditions):
        return self.dao.delete(conditions)

    def transform(self, o, code=None):
        return self.dao.transform(o)

    def tree(self, root, conditions=None):
        return self.dao.tree(conditions, root)



    # def values(self, props, conditions, orders=None):
    #     pass
    #
    # def group(self, props, conditions, orders=None):
    #     pass
    #
    # def count(self, props, conditions):
    #     pass




from pymongo.results import InsertOneResult
from pymongo.synchronous.cursor import Cursor

from .engine import DaoEngine
from .model_mon import MonModel
from .query_mon import MonQuery
from ..dao_abstract import AbstractDao


class MonDao(AbstractDao):
    def __init__(self, entity):
        self.entity = entity
        engine = DaoEngine().mon
        self.model:MonModel = engine.get_model(entity)
        self.objects = self.model.manager

    def get_model(self):
        return self.model

    def save_one(self, item):
        oid, item = self.model.detach_item(item)
        if oid:
            self.objects.update_one({'_id': oid}, {'$set': item})
        else:
            bean: InsertOneResult = self.objects.insert_one(item)
            oid = bean.inserted_id
            self.objects.update_one({'_id': oid}, {'$set': {'sort': str(oid)}})
        return self.objects.find_one({'_id': oid})

    def delete_one(self, _id):
        oid = self.model.object_id(_id)
        flag = False
        if oid:
            self.objects.delete_one({'_id': oid})
            flag = True
        return flag

    def find_one(self, _id):
        oid = self.model.object_id(_id)
        return self.objects.find_one({'_id': oid}) if oid else None

    def find(self, conditions=None, orders=None, size=0, page=1):
        query = MonQuery(conditions, orders)
        skip = (page - 1) * size
        total = self.objects.count_documents(query.finder)
        cursor = self.objects.find(query.finder, query.orders).skip(skip)
        if size:
            cursor = cursor.limit(size)
        return cursor, total

    def update(self, template, conditions):
        if template and len(template.keys()):
            oid, item = self.model.detach_item(template)
            query = MonQuery(conditions)
            self.objects.update_many(query.finder, {'$set': item})
            return True
        return False

    def delete(self, conditions):
        query = MonQuery(conditions)
        self.objects.delete_many(query.finder)

    def transform(self, o, code=None):
        if isinstance(o, Cursor):
            return [__to_item__(doc) for doc in o]
        else:
            return __to_item__(o)

    def tree(self, root, conditions=None):
        pass


def __to_item__(o):
    o['id'] = str(o['_id'])
    del o['_id']
    return o


import json

from .handler import save_many_handler,delete_many_handler
from ...channels.executer import execute_channel
from ...channels.sender import ValarSocketSender
from ...core.dao.dao_base import Dao
from ...core.meta.meta_orm import OrmMeta
from ...core.response import ValarResponse

async def save_many(request):
    sender = ValarSocketSender(request)
    await execute_channel(save_many_handler, sender)
    return ValarResponse(True)

async def delete_many(request):
    sender = ValarSocketSender(request)
    await execute_channel(delete_many_handler, sender)
    return ValarResponse(True)

def save_one (request,db, entity):
    item = json.loads(request.body)
    dao = Dao(entity, db)
    bean = dao.save_one(item)
    item = dao.transform(bean)
    return ValarResponse(item)

def delete_one(request, db, entity):
    body = json.loads(request.body)
    _id = body['id']
    dao = Dao(entity, db)
    flag = dao.delete_one(_id)
    return ValarResponse(flag)

def find_one(request, db, entity):
    body = json.loads(request.body)
    _id = body['id']
    dao = Dao(entity, db)
    bean = dao.find_one(_id)
    item = dao.transform(bean)
    return ValarResponse(item)

def find (request,db, entity):
    conditions = json.loads(request.body)
    dao = Dao(entity, db)
    results, _ = dao.find(conditions)
    results = dao.transform(results)
    return ValarResponse(results)

def update(request, db, entity):
    body = json.loads(request.body)
    conditions = body.get('conditions',[])
    template =body.get('template')
    dao = Dao(entity, db)
    flag = dao.update(template, conditions)
    return ValarResponse(flag)

def table(request, db, entity):
    body = json.loads(request.body)
    conditions = body.get('conditions', [])
    orders = body.get('orders')
    size = body.get('size')
    page = body.get('page')
    dao = Dao(entity, db)
    results, total = dao.find(conditions, orders, size, page)
    results = dao.transform(results)
    return ValarResponse({
        "results":results,
        "total": total
    })

def tree(request, db, entity):
    body = json.loads(request.body)
    conditions = body.get('conditions', [])
    root = body.get('root')
    dao = Dao(entity, db)
    query_set = dao.tree(root, conditions)
    results = dao.transform(query_set)
    return ValarResponse({
        "results": results,
        "root": root
    })


def meta_view(request, db, entity):
    body = json.loads(request.body)
    code = body.get('code')
    _meta = OrmMeta(entity, code)
    _view = _meta.get_view()
    return ValarResponse(_view)


from abc import ABC, abstractmethod

class AbstractDao(ABC):


    @abstractmethod
    def get_model(self):
        pass

    @abstractmethod
    def save_one(self, item):
        pass

    @abstractmethod
    def delete_one(self, _id):
        pass

    @abstractmethod
    def find_one(self, _id):
        pass

    @abstractmethod
    def find(self, conditions=None, orders=None, size=0, page=1):
        pass

    @abstractmethod
    def update(self, template, conditions):
        pass

    @abstractmethod
    def delete(self, conditions):
        pass

    @abstractmethod
    def transform(self, o, code=None):
        pass

    @abstractmethod
    def tree(self, root, conditions=None):
        pass

    def search(self, includes=None, excludes=None, orders=None):
        conditions = [{
            "includes": includes or {},
            "excludes": excludes or {}
        }]
        query_set, _ = self.find(conditions, orders)
        return query_set

    # @abstractmethod
    # def values(self, props, conditions, orders=None):
    #     pass
    #
    # @abstractmethod
    # def group(self, props, conditions, orders=None):
    #     pass
    #
    # @abstractmethod
    # def count(self, props, conditions):
    #     pass






import copy

from .defaults.frame_defaults import meta_field_tool, meta_field_domain

from ...data.models import MetaFieldDomain, MetaFieldTool


def init_meta_frame():
    from ..dao.dao_orm import OrmDao
    tool_dao = OrmDao('data.MetaFieldTool')
    domain_dao = OrmDao('data.MetaFieldDomain')
    mapping = {}
    for item in meta_field_tool:
        _id, code = item['id'], item['code']
        item.update({"saved": True})
        if item['isLeaf']:
            mapping[code] = _id
        tool_dao.save_one(item)

    for row in meta_field_domain:
        default_id, tools = row['default_id'], row['tools']
        print(mapping[default_id])
        _row = copy.deepcopy(row)
        _row.update({
            'default_id': mapping[default_id],
            'tools': [mapping[tool] for tool in tools],
            "saved": True
        })
        domain_dao.save_one(_row)
    return MetaFieldDomain.objects.all()



import json
from io import BytesIO

from django.conf import settings
from minio import Minio
from urllib3 import  BaseHTTPResponse

from ..singleton_meta import SingletonMeta

class MinioEngine(metaclass=SingletonMeta):

    def __int__(self):
        self.client = Minio(**settings.MINIO_SETTINGS)

    def __load_bucket__(self, bucket_name):
        if not self.client.bucket_exists(bucket_name):
            self.client.make_bucket(bucket_name)
            policy = __generate_policy__(bucket_name)
            self.client.set_bucket_policy(bucket_name, policy)

    @staticmethod
    def get_object_name(_id, prop, file_name):
        return f"{_id}-{prop}-{file_name}"

    @staticmethod
    def get_bucket_name(entity):
        name = f'{settings.BASE_DIR.name}.{entity}'
        return name.replace('_', '-').lower()

    def upload(self, bucket_name, object_name, _bytes):
        self.__load_bucket__(bucket_name)
        file_data = BytesIO(_bytes)
        file_size = len(_bytes)  # file.siz
        self.client.put_object(
            bucket_name=bucket_name,
            object_name=object_name,
            data=file_data,
            length=file_size
        )
        return f'{bucket_name}/{object_name}'

    def remove_path(self, path):
        bucket_name, object_name = path.split('/')
        self.remove(bucket_name, object_name)

    def remove(self, bucket_name, object_name):
        self.client.remove_object(
            bucket_name=bucket_name,
            object_name=object_name
        )

    def read(self, bucket_name, object_name) -> BytesIO:
        ret: BaseHTTPResponse = self.client.get_object(bucket_name=bucket_name, object_name=object_name)
        return BytesIO(ret.read())


def __generate_policy__(bucket_name):
    return json.dumps({
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": "s3:GetBucketLocation",
                "Resource": f"arn:aws:s3:::{bucket_name}"
            },
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": "s3:ListBucket",
                "Resource": f"arn:aws:s3:::{bucket_name}"
            },
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": "s3:GetObject",
                "Resource": f"arn:aws:s3:::{bucket_name}/*"
            },
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": "s3:PutObject",
                "Resource": f"arn:aws:s3:::{bucket_name}/*"
            }
        ]})


from deepmerge import always_merger

from .defaults.view_defaults import meta_view_default_values
from ...core.dao.dao_orm import OrmDao
from ...data.models import MetaView, Meta


class OrmMeta:

    def __init__(self, entity, code):
        self.entity = entity
        self.dao = OrmDao(entity)
        self.code = code or 'default'


        # load_meta
        meta_dao = OrmDao('data.Meta')
        _meta = {"entity": entity}
        meta: Meta = meta_dao.search(_meta).first()
        if meta is None:
            _meta.update({"name": self.dao.model.name, 'saved': True})
            meta = meta_dao.save_one(_meta)
        self.meta = meta


        # load_view
        view_dao = OrmDao('data.MetaView')
        _view = {"code": self.code, "meta_id": meta.id}
        view: MetaView = view_dao.search(_view).first()
        if view is None:
            _view.update({"name": self.code.upper(), "saved": True})
            view = view_dao.save_one(_view)
            self.__init_view__(view)
        self.view = view

        # load_fields
        if view.metafield_set.count() == 0:
            field_dao = OrmDao('data.MetaField')
            _fields = self.dao.model.initial_fields(code)
            for _field in _fields:
                _field.update({'view_id':view.id, "saved": True })
                field_dao.save_one(_field)


    def get_view(self):
        _view = self.view.json()
        meta = self.meta.json()
        name, entity = meta['name'], meta['entity']
        fields = self.view.metafield_set.all().order_by('-sort')
        _fields = {field.prop: field.json(entity=entity, code=self.code, db='orm') for field in fields}
        _view.update({
            '$db': 'orm',
            '$entity': entity,
            '$code': self.code,
            '$meta_name': name,
            '$is_tree': self.dao.model.is_tree,
            '$fields': _fields
        })
        return _view

    def __init_view__(self, view: MetaView):
        default_view = meta_view_default_values.get(self.entity, {})
        default_values = default_view.get('__init__',{})
        code_values = default_view.get(self.code, {})
        values = always_merger.merge(default_values, code_values)
        if len(values) > 0:
            for key, value in values.items():
                setattr(view, key, value)
            view.save()


from django.apps import apps

from .model_orm import OrmModel
from ..valar_models import VModel


class OrmEngine:

    def __init__(self):
        mapping = {}
        for model in apps.get_models():
            if issubclass(model, VModel):
                path, name = model.__module__, model.__name__
                prefix = 'src.valar.' if path.startswith('src') else 'valar.'
                app = path.replace('.models', '').replace(prefix, '')
                entity = '%s.%s' % (app, name)
                mapping[entity] = model
        self.mapping = mapping

    def get_mapping(self)->dict:
        return self.mapping

    def get_model(self,entity)->OrmModel:
        mod = self.mapping.get(entity)
        return OrmModel(mod, entity)



from django.db.models import QuerySet

from ..valar_models import VModel
from ...core.meta.field_orm import OrmField
from ...data.models import MetaField


class OrmUtils:
    def __init__(self):
        self.multiple_domains = ['ManyToOneRel', 'ManyToManyField', 'ManyToManyRel']
        self.referred_domains = [*self.multiple_domains, 'OneToOneRel', 'OneToOneField', 'ForeignKey']
        self.omit_field_props = ['create_time', 'modify_time', 'saved', 'sort']
        self.data_props_formatting = {'DateField': '%Y-%m-%d', 'DateTimeField': '%Y-%m-%d %H:%M:%S', 'TimeField': '%H:%M:%S'}


    @staticmethod
    def json(bean: VModel):
        pass

    @staticmethod
    def custom_props(entity, code='default'):
        field_set = MetaField.objects.filter(view__code=code, view__meta__entity=entity, domain='Custom').values('prop')
        return [item['prop'] for item in field_set if item['prop']]


    def linkage(self, field, query_set: QuerySet, mapping):
        model_field = field.model_field
        prop = model_field.name
        multiple = field.domain in self.multiple_domains

        # 获取级联关系的键索引
        ref_prop = f'{prop}__id'
        edges = query_set.exclude(**{f'{ref_prop}__isnull': True}).values('id', ref_prop)
        if multiple:
            related_primary_keys = set()
            results_mapping = {}
            for edge in edges:
                _id, rid = edge['id'], edge[ref_prop]
                related_primary_keys.add(rid)
                array = results_mapping.get(_id, [])
                array.append(rid)
                results_mapping[_id] = array
        else:
            results_mapping = {row['id']: row[ref_prop] for row in edges if row[ref_prop]}
            related_primary_keys = set(results_mapping.values())

        # 获取级联关系从属方的数据
        related_model = model_field.related_model
        related_fields = related_model._meta.get_fields()
        related_props = self.__get_related_props__(related_fields)
        related_values = list(related_model.objects.filter(id__in=related_primary_keys).values(*related_props))
        self.date_values(related_fields, related_values)
        related_mapping = {item['id']: item for item in related_values}

        # 将从属方的数据绑定在主数据上
        for _id in mapping:
            row = mapping[_id]
            if multiple:
                keys = results_mapping.get(_id, [])
                items = [related_mapping[pid] for pid in keys]
                row[prop] = keys
                row[f'{prop}_set'] = items
            else:
                key = results_mapping.get(_id)
                item = related_mapping.get(key) if key else None
                row[prop] = item
                row[f'{prop}_id'] = key

    def __get_related_props__(self,fields):
        def fun(field): return type(field).__name__ not in self.referred_domains and field.name not in self.omit_field_props
        return [field.name for field in fields if fun(field)]

    def date_values(self, fields, values):
        date_props_mapping = {}
        for field in fields:
            if isinstance(field, OrmField):
                prop = field.prop
                domain = field.domain
            else:
                prop = field.name
                domain = type(field).__name__
            if domain in self.data_props_formatting.keys():
                date_props_mapping[prop] = self.data_props_formatting[domain]
        for row in values:
            for prop, formating in date_props_mapping.items():
                if row.get(prop):
                    row[prop] = row[prop].strftime(formating)


from django.urls import path

from ..channels.views import handel_channel
from .views.rest import save_many,delete_many,save_one,delete_one,find_one,find,update,table, tree, meta_view

urlpatterns = [
    path('socket/<str:handler>', handel_channel),
    path('save_many', save_many),
    path('delete_many', delete_many),
    path('<str:db>/<str:entity>/save_one', save_one),
    path('<str:db>/<str:entity>/delete_one', delete_one),
    path('<str:db>/<str:entity>/find_one', find_one),
    path('<str:db>/<str:entity>/find', find),
    path('<str:db>/<str:entity>/update', update),
    path('<str:db>/<str:entity>/table', table),
    path('<str:db>/<str:entity>/tree', tree),
    path('<str:db>/<str:entity>/meta_view', meta_view),
]